import unittest
import requests
from time import sleep
from datetime import datetime, timedelta
from pyrecard.subscription import plan, customer, subscription, payment, coupon, webhook
from pyrecard.utils.factory import url_factory
from pyrecard.test import mock
from os import environ


class PlanTestCase(unittest.TestCase):

    def setUp(self):
        self.data = mock.plan_data()

    def test_wirecard_sandbox_must_be_accessible(self):
        response = requests.get(url_factory())
        self.assertEqual(401, response.status_code)
        self.assertEqual('Token or Key are invalids', response.json()['ERROR'])

    def test_set_response_timeout_must_be_returned(self):
        environ['PYRECARD_REQUEST_TIMEOUT'] = '0.001'
        response = plan.create(self.data)
        self.assertEqual(408, response.status_code)
        self.assertTrue('error' in response.json())

    def test_plan_with_valid_data_should_be_created(self):
        response = plan.create(self.data)
        self.assertEqual(response.status_code, 201)

    def test_plan_with_invalid_data_should_not_be_created(self):
        self.data['code'] = ''
        response = plan.create(self.data)
        self.assertEqual(response.status_code, 400)
        self.assertTrue('errors' in response.json())

    def test_plan_should_be_changed(self):
        plan.create(self.data)
        old_amount = self.data['amount']
        self.data['amount'] = 1000
        code = self.data.pop('code')
        plan.alter(code, self.data)
        response = plan.fetch(code)
        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.json().get('amount'), old_amount)

    def test_all_plans_must_be_returned(self):
        response = plan.fetch_all()
        self.assertEqual(response.status_code, 200)
        self.assertTrue('plans' in response.json())
        self.assertGreater(len(response.json()['plans']), 1)

    def test_plan_must_be_inactivated_and_reactivated(self):
        plan.create(self.data)
        response = plan.fetch(self.data['code'])
        self.assertEqual(response.json()['status'], 'ACTIVE')
        plan.inactivate(self.data['code'])
        response = plan.fetch(self.data['code'])
        self.assertNotEqual(response.json()['status'], 'ACTIVE')
        plan.activate(self.data['code'])
        response = plan.fetch(self.data['code'])
        self.assertEqual(response.json()['status'], 'ACTIVE')

    def tearDown(self):
        environ['PYRECARD_REQUEST_TIMEOUT'] = '5'


class CustomersTestCase(unittest.TestCase):

    def setUp(self):
        self.data = mock.customer_data()

    def test_customer_with_valid_data_must_be_created(self):
        response = customer.create(self.data)
        self.assertEqual(response.status_code, 201)

    def test_customer_must_be_changed(self):
        customer.create(self.data)
        old_street_name = self.data['address']['street']
        self.data['address']['street'] = 'New Street Name'
        customer.alter(self.data['code'], self.data)
        response = customer.fetch(self.data['code'])
        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.json()['address']['street'], old_street_name)

    def test_all_customers_must_be_returned(self):
        response = customer.fetch_all()
        self.assertEqual(response.status_code, 200)
        self.assertTrue('customers' in response.json())
        self.assertGreater(len(response.json()['customers']), 1)

    def test_credit_card_must_be_changed(self):
        customer.create(self.data)
        credit_card = self.data['billing_info']
        credit_card['credit_card']['number'] = '5555666677778884'
        response = customer.change_card(self.data['code'], credit_card)
        self.assertEqual(response.status_code, 200)
        self.assertTrue('message' in response.json())


class SubscriptionTestCase(unittest.TestCase):

    def setUp(self):
        self.data = mock.subscription_data()

    def test_raise_error_if_payment_method_is_invalid(self):
        response = subscription.set_payment_method(self.data['code'], 'INVALID_METHOD')
        self.assertEqual(response.status_code, 400)
        self.assertTrue('error' in response.json())

    def test_raise_error_if_subscription_status_is_invalid(self):
        response = subscription.set_status(self.data['code'], 'INVALID_STATUS')
        self.assertTrue(response.status_code, 400)
        self.assertTrue('error' in response.json())

    def test_signature_must_be_created(self):
        response = subscription.create(self.data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()['message'], 'Assinatura criada com sucesso')

    def test_signature_must_be_created_with_new_customer(self):
        self.data['customer'] = mock.customer_data()
        response = subscription.create(self.data, True)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()['message'], 'Assinatura criada com sucesso')

    def test_subscription_details_must_be_returned(self):
        subscription.create(self.data)
        response = subscription.fetch(self.data['code'])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()['code'], self.data['code'])

    def test_all_subscriptions_must_be_returned(self):
        response = subscription.fetch_all()
        self.assertEqual(response.status_code, 200)
        self.assertTrue('subscriptions' in response.json())
        self.assertGreater(len(response.json()['subscriptions']), 1)

    def test_subscription_status_must_changed(self):
        subscription.create(self.data)
        response = subscription.set_status(self.data['code'], 'suspend')
        self.assertEqual(response.status_code, 200)
        response = subscription.set_status(self.data['code'], 'activate')
        self.assertEqual(response.status_code, 200)
        response = subscription.set_status(self.data['code'], 'cancel')
        self.assertEqual(response.status_code, 200)

    def test_subscription_data_must_be_changed(self):
        subscription.create(self.data)
        data = mock.plan_data()
        plan.create(data)
        self.data['plan']['code'] = data['code']
        code = self.data.pop('code')
        del self.data['customer']
        response = subscription.alter(code, self.data)
        self.assertEqual(response.status_code, 200)

    def test_subscription_payment_method_must_be_changed(self):
        subscription.create(self.data)
        response = subscription.set_payment_method(self.data['code'], 'BOLETO')
        self.assertEqual(response.status_code, 200)

    def test_all_invoices_must_be_returned(self):
        response = subscription.fetch_all_invoices(self.data['code'])
        self.assertEqual(response.status_code, 200)
        self.assertTrue('invoices' in response.json())


class InvoicesTestCase(unittest.TestCase):

    def setUp(self):
        self.data = mock.subscription_data()
        self.customer = mock.customer_data()

    def test_invoice_must_be_returned(self):
        sub = subscription.create(self.data)
        response = payment.fetch_invoice(sub.json()['invoice']['id'])
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.json(), dict)

    def test_all_invoice_payments_must_be_returned(self):
        sub = subscription.create(self.data)
        response = payment.fetch_invoice_payments(sub.json()['invoice']['id'])
        self.assertEqual(response.status_code, 200)
        self.assertTrue('payments' in response.json())

    def test_payment_details_must_be_returned(self):
        sub = subscription.create(self.data)
        invoices = payment.fetch_invoice_payments(sub.json()['invoice']['id'])
        response = payment.payment_details(invoices.json()['payments'][0]['id'])
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.json(), dict)

    def test_payment_attempt_must_be_made(self):
        self.customer['billing_info']['credit_card']['number'] = '5168441223630339'
        customer.change_card(self.data['customer']['code'], self.customer['billing_info'])
        sub = subscription.create(self.data).json()
        invoice_id, sub_code = sub['invoice']['id'], sub['code']
        sleep(14)
        sub = subscription.fetch(sub_code)
        self.assertEqual('Atrasada', sub.json()['invoice']['status']['description'])
        self.customer['billing_info']['credit_card']['number'] = '4111111111111111'
        customer.change_card(self.data['customer']['code'], self.customer['billing_info'])
        response = payment.retry_invoice_payment(invoice_id)
        self.assertTrue(response.status_code, 200)
        sub = subscription.fetch(sub_code)
        self.assertTrue(sub.json()['invoice']['status']['description'], 'Pago')

    def teste_bank_slip(self):
        self.data['payment_method'] = 'BOLETO'
        sub = subscription.create(self.data)
        date = datetime.now() + timedelta(5)
        self.assertTrue('boleto' in sub.json()['_links'])
        response = payment.generate_bank_slip(
            sub.json()['invoice']['id'],
            {"year": date.year, "month": date.month, "day": date.day},
        )
        self.assertTrue(response.status_code, 200)
        self.assertNotEqual(sub.json()['_links']['boleto'], response.json()['_links']['boleto'])


class CouponTestCase(unittest.TestCase):

    def setUp(self):
        self.data = mock.coupon_data()
        self.json = {'coupon': {'code': self.data['code']}}

    def test_returns_error_if_status_is_not_valid(self):
        response = coupon.set_status(self.data['code'], 'invalid-status')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()['error'], 'The coupon status must be "active" or "inactive"')

    def test_coupon_must_be_created(self):
        response = coupon.create(self.data)
        self.assertTrue(response.ok)
        self.assertTrue("creation_date" in response.json())

    def test_coupon_must_be_applied(self):
        sub_data = mock.subscription_data()
        sub = subscription.create(sub_data)
        self.assertFalse('coupon' in sub.json())
        self.assertTrue(coupon.create(self.data).ok)
        response = coupon.apply(sub_data['code'], self.data['code'])
        self.assertTrue(response.ok)
        sub = subscription.fetch(sub_data['code'])
        self.assertTrue('coupon' in sub.json())

    def test_coupon_must_be_returned(self):
        coupon.create(self.data)
        response = coupon.fetch(self.data['code'])
        self.assertTrue(response.ok)
        self.assertTrue('creation_date' in response.json())

    def test_all_coupons_must_be_returned(self):
        response = coupon.fetch_all()
        self.assertTrue(response.ok)
        self.assertTrue('coupons' in response.json())
        self.assertIsInstance(response.json()['coupons'], list)

    def test_coupon_activation_must_change(self):
        data = coupon.create(self.data)
        self.assertEqual(data.json()['status'], 'ACTIVE')
        response = coupon.set_status(self.data['code'], 'inactive')
        self.assertTrue(response.ok)
        self.assertNotEqual(response.json()['status'], 'ACTIVE')
        response = coupon.set_status(self.data['code'], 'active')
        self.assertEqual(response.json()['status'], 'ACTIVE')

    def test_coupon_must_be_removed_from_subscription(self):
        sub_data = mock.subscription_data()
        sub = subscription.create(sub_data)
        self.assertFalse('coupon' in sub.json())
        self.assertTrue(coupon.create(self.data).ok)
        self.assertTrue(coupon.remove(sub_data['code']).ok)
        response = subscription.fetch(sub_data['code'])
        self.assertFalse('coupon' in response.json())


class WebhookTestCase(unittest.TestCase):

    def test_webhook_settings_must_be_set(self):
        self.assertTrue(
            webhook.preferences('http://exemploldeurl.com.br/assinaturas', True, True).ok,
        )



from pyrecard.utils.pyrequest import pyrequest


def fetch_invoice(invoice_id):
    return pyrequest('GET', f'/invoices/{invoice_id}')


def fetch_invoice_payments(invoice_code):
    return pyrequest('GET', f'/invoices/{invoice_code}/payments')


def payment_details(payment_code):
    return pyrequest('GET', f'/payments/{payment_code}')


def retry_invoice_payment(invoice_code):
    return pyrequest('POST', f'/invoices/{invoice_code}/retry')


def generate_bank_slip(invoice_code, json={}):
    return pyrequest('POST', f'/invoices/{invoice_code}/boletos', json)



from random import choices
from datetime import datetime
from pyrecard.subscription import plan, customer


def plan_data():
    '''
       The following data has been taken and adapted from
       https://dev.wirecard.com.br/v1.5/reference#criar-plano
    '''
    code = ''.join(choices('ABCDEF1234567890', k=10))
    return {
        'code': code,
        'name': f'Plano {code}',
        'description': f'Descrição do Plano {code}',
        'amount': 990,
        'setup_fee': 500,
        'max_qty': 1,
        'interval': {'length': 1, 'unit': 'MONTH'},
        'billing_cycles': 12,
        'payment_method': 'ALL'
    }


def customer_data():
    '''
       The following data has been taken and adapted from
       https://dev.wirecard.com.br/v1.5/reference#criar-assinante
    '''
    code = ''.join(choices('ABCDEF1234567890', k=10))
    return {
        'code': code,
        'email': f'cliente{code}@provedor.com',
        'fullname': f'Usuário {code}',
        'cpf': '36598211018',  # cpf obtained in https://www.4devs.com.br/gerador_de_cpf
        'phone_area_code': '11',
        'phone_number': '934343434',
        'birthdate_day': '26',
        'birthdate_month': '04',
        'birthdate_year': '1980',
        'address': {
            'street': 'Street Name',
            'number': '100',
            'complement': '',
            'district': 'Neighboor Name',
            'city': 'City Name',
            'state': 'MG',
            'country': 'BRA',
            'zipcode': '00000000',
        },
        'billing_info': {
            'credit_card': {
                'holder_name': 'Test User',
                'number': '4111111111111111',
                'expiration_month': '06',
                'expiration_year': str(datetime.now().year+2)[2:]
            }
        }
    }


def subscription_data():
    '''
       The following data has been taken and adapted from
       https://dev.wirecard.com.br/v1.5/reference#assinaturas
    '''
    mocked_plan = plan_data()
    mocked_customer = customer_data()
    plan.create(mocked_plan)
    customer.create(mocked_customer, new_vault=True)
    return {
        'code': ''.join(choices('ABCDEF1234567890', k=10)),
        'amount': 9990,
        'payment_method': 'CREDIT_CARD',
        'plan': {'code': mocked_plan['code']},
        'customer': {'code': mocked_customer['code']},
    }


def coupon_data():
    '''
       The following data has been taken and adapted from
       https://dev.wirecard.com.br/v1.5/reference#criar-coupon
    '''
    code = ''.join(choices('ABCDEF1234567890', k=10))
    return {
        'code': f'coupon-{code}',
        'name': f'Coupon {code}',
        'description': 'My new coupon',
        'discount': {
            'value': 10000,
            'type': 'percent'
        },
        'status': 'active',
        'duration': {
            'type': 'repeating',
            'occurrences': 12
        },
        'max_redemptions': 1000,
        'expiration_date': {
            'year': datetime.now().year,
            'month': datetime.now().month,
            'day': datetime.now().day,
        }
    }



__title__ = 'pyrecard'
__description__ = 'A Python lib for brazilian Wirecard payment gateway.'
__url__ = 'https://github.com/DiegoMagg/pyrecard'
__version__ = '1.0.0'
__author__ = 'Diego Magalhães'
__author_email__ = 'dmlmagal@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2020 Diego Magalhães'



from pyrecard.utils.pyrequest import pyrequest


def create(json, new_vault=False):
    return pyrequest('POST', f'/customers?new_vault={new_vault}', json)


def alter(customer_code, json):
    return pyrequest('PUT', f'/customers/{customer_code}', json)


def fetch(customer_code):
    return pyrequest('GET', f'/customers/{customer_code}')


def fetch_all():
    return pyrequest('GET', '/customers')


def change_card(customer_code, json):
    return pyrequest('PUT', f'/customers/{customer_code}/billing_infos', json)



from pyrecard.utils.pyrequest import pyrequest
from pyrecard.subscription.subscription import alter
from pyrecard.utils.factory import response_factory


def create(json):
    return pyrequest('POST', '/coupons', json)


def apply(subscription_code, coupon_code):
    return alter(subscription_code, {'coupon': {'code': coupon_code}})


def fetch(coupon_code):
    return pyrequest('GET', f'/coupons/{coupon_code}')


def fetch_all():
    return pyrequest('GET', '/coupons')


def set_status(coupon_code, status):
    if str(status).lower() not in ('active', 'inactive'):
        return response_factory(400, {'error': 'The coupon status must be "active" or "inactive"'})
    return pyrequest('PUT', f'/coupons/{coupon_code}/{status}')


def remove(coupon_code):
    return pyrequest('DELETE', f'/subscriptions/{coupon_code}/coupon')



from pyrecard.utils.pyrequest import pyrequest


def preferences(url, email_merchant_enabled, customer_enabled):
    data = {
        'notification': {
            'webhook': {'url': url},
            'email': {
                'merchant': {'enabled': email_merchant_enabled},
                'customer': {'enabled': customer_enabled},
            }
        }
    }
    return pyrequest('POST', '/users/preferences', data)



from pyrecard.utils.pyrequest import pyrequest
from pyrecard.utils.factory import response_factory


def create(json, new_customer=False):
    return pyrequest('POST', f'/subscriptions?new_customer={new_customer}', json)


def fetch(subscription_code):
    return pyrequest('GET', f'/subscriptions/{subscription_code}')


def fetch_all():
    return pyrequest('GET', '/subscriptions')


def set_status(subscription_code, status):
    message = 'The subscription status must be "suspend", "activate" or "cancel"'
    if str(status).lower() not in ('suspend', 'activate', 'cancel'):
        return response_factory(400, {'error': message})
    return pyrequest('PUT', f'/subscriptions/{subscription_code}/{status}')


def alter(subscription_code, json):
    return pyrequest('PUT', f'/subscriptions/{subscription_code}', json)


def set_payment_method(subscription_code, method):
    if str(method).upper() not in ('CREDIT_CARD', 'BOLETO'):
        return response_factory(400, {'error': 'The payment method must be "CREDIT_CARD" or "BOLETO"'})
    return pyrequest(
        'PUT',
        f'/subscriptions/{subscription_code}/change_payment_method',
        json={'payment_method': method.upper()},
    )


def fetch_all_invoices(subscription_code):
    return pyrequest('GET', f'/subscriptions/{subscription_code}/invoices')



from pyrecard.utils.pyrequest import pyrequest


def create(json):
    return pyrequest('POST', '/plans', json)


def alter(plan_code, json):
    return pyrequest('PUT', f'/plans/{plan_code}', json)


def activate(plan_code):
    return pyrequest('PUT', f'/plans/{plan_code}/activate')


def inactivate(plan_code):
    return pyrequest('PUT', f'/plans/{plan_code}/inactivate')


def fetch(plan_code):
    return pyrequest('GET', f'/plans/{plan_code}')


def fetch_all():
    return pyrequest('GET', '/plans')



import unittest
from os import environ
from pyrecard.exceptions import MissingKey, InvalidKey
from pyrecard.utils.factory import b64_factory, header_factory


class FactoryTestCase(unittest.TestCase):

    def test_raise_error_if_not_key(self):
        with self.assertRaises(MissingKey):
            b64_factory(None)

    def test_raise_error_if_key_is_invalid(self):
        with self.assertRaises(InvalidKey):
            b64_factory('1010:ABAB')

    def test_generate_b64_object_with_a_valid_key(self):
        header = b64_factory(environ.get('SANDBOX_KEY'))
        self.assertTrue('Authorization' in header)
        self.assertTrue('Basic' in header['Authorization'])

    def test_header_with_authorization_must_be_returned_in_sandbox(self):
        '''
           if the environment variable "PYRECARD_ENV" isn't to "production"
           then the key used will be "SANDBOX_KEY" will be used.
        '''
        headers = header_factory()
        self.assertTrue('User-Agent' in headers)
        self.assertTrue('Authorization' in headers)

    def test_header_with_authorization_must_be_returned_in_production(self):
        '''
           if the environment variable "PYRECARD_ENV" is set to "production"
           then the key used will be "PRODUCTION_KEY".
        '''
        environ['PRODUCTION_KEY'] = environ.get('SANDBOX_KEY')
        environ['PYRECARD_ENV'] = 'production'
        headers = header_factory()
        self.assertTrue('User-Agent' in headers)
        self.assertTrue('Authorization' in headers)

    def tearDown(self):
        environ['PYRECARD_ENV'] = ''



'''Module containing exceptions raised by pyrecard.'''


class PyrecardException(Exception):
    '''Base class for all exceptions raised by pyrecard.'''

    pass


class MissingKey(PyrecardException):
    '''Raised when wirecard key is missing'''

    pass


class InvalidKey(PyrecardException):
    '''Raised when wirecard key is invalid'''

    pass



import setuptools
from pyrecard.__version__ import (
    __version__,
    __author__,
    __author_email__,
    __description__,
    __url__,
)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyrecard",
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=__url__,
    packages=setuptools.find_packages(),
    install_requires=['requests'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from os import environ
from requests.models import Response
from re import match
from base64 import b64encode
from pyrecard.exceptions import MissingKey, InvalidKey
from pyrecard.__version__ import __version__
from json import dumps


def url_factory(path=''):
    url = 'https://{}.moip.com.br/assinaturas/v1{}'
    production = environ.get('PYRECARD_ENV') == 'production'
    return url.format('api', path) if production else url.format('sandbox', path)


def response_factory(status_code, json):
    response = Response()
    response.status_code = status_code
    response._content = f'{dumps(json)}'.encode('UTF-8')
    return response


def b64_factory(key=None):
    if not key:
        raise MissingKey('Missing auth key')
    if not match(r'\w{32}:\w{40}', key):
        raise InvalidKey('Invalid key')
    return {'Authorization': f'Basic {str(b64encode(key.encode("UTF-8")), "UTF-8")}'}


def header_factory():
    headers = {'User-Agent': f'Pyrecard - {__version__}', 'Content-Type': 'application/json'}
    if environ.get('PYRECARD_ENV', '').lower() == 'production':
        return {**headers, **b64_factory(environ.get('PRODUCTION_KEY'))}
    return {**headers, **b64_factory(environ.get('SANDBOX_KEY'))}



from requests import request
from requests.exceptions import ConnectTimeout, ReadTimeout
from os import environ
from pyrecard.utils.factory import header_factory, url_factory, response_factory


def pyrequest(method, path, json={}):
    try:
        return request(
            method.upper(),
            url_factory(path),
            headers=header_factory(),
            json=json,
            timeout=float(environ.get('PYRECARD_REQUEST_TIMEOUT', 5)),
        )
    except (ConnectTimeout, ReadTimeout):
        return response_factory(408, {'error': f'Connection to {url_factory()} timed out'})
